/*
 * (C) Copyright Bora Software Inc. 2002, 2003 - All Rights Reserved.
 *
 * $Id: CommandShell.java,v 1.18 2002/12/08 00:11:02 kboo Exp $
 */

package com.bs.xdbms.cli;

import java.io.IOException;

import com.bs.xdbms.util.Constants;

import org.xmldb.api.modules.CollectionManagementService;
import org.xmldb.api.base.Collection;
import org.xmldb.api.base.Database;
import org.xmldb.api.DatabaseManager;

public class CommandShell {
  private final String VERSION = "0.1";
  private final String PROMPT = "myxdb>";

  private Collection currentCollection;

  private String[] help = {
    "help             (?)  Display this help.",
    "dir                   Displays the files in the current directory.",
    "add doc_name     (a)  Add an XML document to the current collection.",
    "list             (l)  List XML documents in the current collection.",
    "update doc_name  (u)  Update an XML document in the collection.",
    "extract doc_name (e)  Retrieve an XML document to the current directory.",
    "delete doc_name  (d)  Remove an XML document from the collection.",
    "show [all]            Show current collection or entire collection names.",
    "lc                    List child collection names.",
    "use col_name          Change current collection to col_name.",
    "create col_name  (c)  Create a child collection.",
    "dc col_name           Delete the collection and its children from the database.",
    "quit             (q)  Quit myxdb."
  };

  public static void main(String[] args) {
    CommandShell shell = new CommandShell();
    shell.run();
  }

  public void run() {
    greeting();
    try {
      currentCollection = getRootCollection();
    } catch (Exception e) {
      e.printStackTrace();
      System.out.println("FATAL ERROR: Failed retrieving the root collection.");
    }

    String commandLine;
    print(PROMPT);
    while(true) {
      try {
        commandLine = getLine(); // get the line input
        //System.out.println("getLine() returned: " + commandLine);
        processCommandLine(commandLine);
        print(PROMPT);
        commandLine = getLine(); // consume PROMPT 
      } catch (IOException e) {
        System.out.println("ERROR: Invalid command.");
        println("Type 'help' or ? for help.\n");
        continue;
      } catch (Exception e) {
        // TBD
        e.printStackTrace();
        continue;
      }
    }
  }

  private Collection getCurrentCollection() {
    return currentCollection;
  }
  
  private void setCurrentCollection(Collection coll) {
    currentCollection = coll;
  }

  private Collection getRootCollection() throws Exception {
    Collection col = null;
    String driver = "com.bs.xdbms.xmldb.base.Database";
    Class c = Class.forName(driver);
     
    Database database = (Database)c.newInstance();
    DatabaseManager.registerDatabase(database);
    col = DatabaseManager.getCollection
      ("xmldb:myxdb://localhost/root","root","");      

    return col;
  }

  private void processCommandLine(String line) throws Exception {
    String commandLine = line.trim();
    if (commandLine.length() == 0)
      return; // user just hit return

    int index = commandLine.indexOf(" ");
    String command;
    String arguments = null;
    if (index == -1) // command without arguments
      command = commandLine;
    else
    {
      command = commandLine.substring(0,index);
      arguments = commandLine.substring(index+1,commandLine.length());
    }

    processCommand(command,arguments);
  }

  private void processCommand(String command, String arguments) 
      throws Exception {
    //TBD
    //System.out.println("Command entered: " + command);
    //System.out.println("Command arguments entered: " + arguments);
    if (command.equalsIgnoreCase("?") 
        || command.equalsIgnoreCase("help"))
      help();
    else if (command.equalsIgnoreCase("a") 
        || command.equalsIgnoreCase("add"))
      add(arguments);
    else if (command.equalsIgnoreCase("l") 
        || command.equalsIgnoreCase("list"))
      list();
    else if (command.equalsIgnoreCase("u") 
        || command.equalsIgnoreCase("update"))
      update(arguments);
    else if (command.equalsIgnoreCase("e") 
        || command.equalsIgnoreCase("extract"))
      get(arguments);
    else if (command.equalsIgnoreCase("d") 
        || command.equalsIgnoreCase("del")    
        || command.equalsIgnoreCase("delete"))
      delete(arguments);
    else if (command.equalsIgnoreCase("dir"))
      dir();
    else if (command.equalsIgnoreCase("show"))
      showCollectionNames(arguments);
    else if (command.equalsIgnoreCase("lc")) 
      listChildCollectionNames();
    else if (command.equalsIgnoreCase("use"))
      useCollection(arguments);
    else if (command.equalsIgnoreCase("c") 
             || command.equalsIgnoreCase("create"))
      createCollection(arguments);
    else if (command.equalsIgnoreCase("dc")) 
      deleteCollection(arguments);
    else if (command.equalsIgnoreCase("q") 
        || command.equalsIgnoreCase("quit"))
      quit();
    else {
      invalidCommand(command);
    }
  }

  private String getLine() throws IOException {
    // TBD - skip leading space characters
    //while (Character.isSpace((char)System.in.read()));

    int c;
    StringBuffer buffer = new StringBuffer();
    while ((c=System.in.read()) != '\n' && c != '\r') 
      buffer.append((char)c);

    return buffer.toString();
  }

  private void println(String s) {
    System.out.println(s);
  }

  private void print(String s) {
    System.out.print(s);
  }

  private void greeting() {
    println("\nMyXDB Command Interpreter version " + VERSION);
    println("Type 'help' or ? for help.\n");
  }

  private void invalidCommand(String cmd) {
    println("ERROR: Invalid command '" + cmd + "'");
    println("Type 'help' or ? for help.\n");
  }

  private void help() {
    println("myxdb commands:");
    for (int i=0; i <help.length; i++) 
      println(help[i]);
    print("\n");
  }

  private void add(String arg) throws Exception {
    Import add = new Import(getCurrentCollection(),arg);
    add.execute();
  }

  private void list() throws Exception {
    ListResources list = new ListResources(getCurrentCollection());
    list.execute();
  }
  
  private void update(String arg) throws Exception {
    Update update = new Update(getCurrentCollection(),arg);
    update.execute();
  }
  
  private void get(String arg) throws Exception {
    Export get = new Export(getCurrentCollection(),arg);
    get.execute();
  }

  private void delete(String arg) throws Exception {
    Delete del = new Delete(getCurrentCollection(),arg);
    del.execute();
  }

  private void dir() throws Exception {
    Dir dir = new Dir();
    dir.execute();
  }

  // Display the absolute path to the current collection.
  // e.g. /root/pub
  // If 'show all' then display the entire collection names.
  private void showCollectionNames(String arguments) throws Exception {
    // arguments may be 'all'
    if (arguments == null) {
      Collection coll = getCurrentCollection();
      String fullName = 
        coll.getProperty(Constants.XDB_COLLECTION_ABSOLUTE_PATH);
      println("Current collection is: " + fullName);
    } else if (arguments.equals("all")) {
      println("Not implemented yet.");
    } else
      invalidCommand("show " + arguments);
  }

  private void listChildCollectionNames() throws Exception {
    String[] children = getCurrentCollection().listChildCollections();
    for (int i=0; i<children.length; i++)
      println("Child collection: " + children[i] + ".");
  }

  private void useCollection(String arguments) throws Exception {
  	// TBD - handle both absolute and relative path patterns.
  	Collection collection;
  	if (arguments.equals("..")) {
  	  collection = getCurrentCollection().getParentCollection();
  	  if (collection == null) // root collection
  	    return;
  	  else
  	    setCurrentCollection(collection);
  	} else if (arguments.equals("."))
  	  ; // do nothing
  	else {
  	  collection = getCurrentCollection().getChildCollection(arguments);
  	  if (collection == null) { // child does not exist
  	    println("The child collection \"" + arguments + "\" does not exist.");
  	    return;
  	  } else                                                
        setCurrentCollection(collection);
  	}
  	showCollectionNames(null);
  }

  private void createCollection(String arguments) throws Exception {
  	CollectionManagementService service =
      (CollectionManagementService)getCurrentCollection().getService("CollectionManagementService","1.0");
    Collection child = service.createCollection(arguments);
    //getCurrentCollection().close(); //TBD - can I reopen later?
    setCurrentCollection(child);
    showCollectionNames(null);
  }

  private void deleteCollection(String arguments) throws Exception {
  	Collection child = getCurrentCollection().getChildCollection(arguments);
  	if (child == null) { // child does not exist
  	  println("The child collection \"" + arguments + "\" does not exist.");
  	  return;
  	}
  	child.close();
  	CollectionManagementService service =
      (CollectionManagementService)getCurrentCollection().getService("CollectionManagementService","1.0");
    service.removeCollection(arguments);
    println("The collection \"" + arguments + "\" has been deleted.");
  }
  
  private void quit() {
    System.exit(0);
  }

}
