/*
 * (C) Copyright Bora Software Inc. 2002, 2003 - All Rights Reserved.
 *
 * $Id: DBUpdater.java,v 1.14 2002/11/16 23:10:30 kboo Exp $
 */
package com.bs.xdbms.persistence;

import java.util.Vector;
import java.sql.*;
import com.bs.xdbms.datamodel.*;

/* 
Usage:
  Connection conn = ...;
  updater = new DBUpdater(conn);
  try {
    updater.insert(xdoc); // for insert
    updater.delete(xdoc); // for delete
  } catch (SQLException e) {
    // handle exception
    updater.rollback();
  }
  updater.commit();
*/

public class DBUpdater {

  // TBD - Performance enhancement.
  // Make the PreparedStatements instance variables where appropriate 
  // for reuse.

  // TBD - review connection synchronization issue for concurrent use
  protected Connection connection;

  private OIDManager oidManager;

  // Add namespaces that have been inserted into the table so far. This
  // is to preveneted the same namespace from being added to the table
  // multiple times because elements may belong to the same namespaces.
  protected Vector namespaces = null;

  public DBUpdater(Connection conn) throws SQLException {
    connection = conn;
    if (connection.getAutoCommit())
      connection.setAutoCommit(false);

    oidManager = new OIDManager(connection);
  }

  public void commit() {
    try {
      connection.commit();
    } catch (SQLException e) {
      // nothing to do
    }
  }

  public void rollback() {
    try {
      connection.rollback();
    } catch (SQLException e) {
      // nothing to do
    }
  }

  public void insert(XDocument xdoc) throws SQLException {
  	namespaces = new Vector();
    long oid = getNewXDocumentOID();
    xdoc.setOID(oid);
    String baseURI = xdoc.getBaseURI();
    String sql = 
      "INSERT INTO X_DOCUMENT (X_DOCUMENT_OID,BASE_URI) VALUES (?,?)";
    PreparedStatement pstmt = connection.prepareStatement(sql);
    pstmt.setLong(1,oid);
    pstmt.setString(2,baseURI);
    pstmt.executeUpdate();
    pstmt.close();

    // Setting relationships with children
    Vector children = xdoc.getChildren();
    XNode child;
    for (int i=0; i<children.size(); i++) {
      child = (XNode)children.elementAt(i);
      if (child.getNodeKind().equals("element")) {
        insert((XElement)child);
      } else if (child.getNodeKind().equals("processing-instruction")) {
        insert((XProcessingInstruction)child);
      } else if (child.getNodeKind().equals("comment")) {
        insert((XComment)child);
      } else
        ; // TBD - Need to raise XModelException ?
      insertPCSequence(xdoc,child);
    } // end for
  }

  public void delete(XDocument xdoc) throws SQLException {
  	namespaces = new Vector();
    long oid = xdoc.getOID();
    //String baseURI = xdoc.getBaseURI();
    String sql = 
      "DELETE FROM X_DOCUMENT WHERE X_DOCUMENT_OID = ?";
    PreparedStatement pstmt = connection.prepareStatement(sql);
    pstmt.setLong(1,oid);
    //pstmt.setString(2,baseURI);
    pstmt.executeUpdate();
    pstmt.close();

    // Delete the children
    Vector children = xdoc.getChildren();
    XNode child;
    for (int i=0; i<children.size(); i++) {
      child = (XNode)children.elementAt(i);
      if (child.getNodeKind().equals("element")) {
        delete((XElement)child);
      } else if (child.getNodeKind().equals("processing-instruction")) {
        delete((XProcessingInstruction)child);
      } else if (child.getNodeKind().equals("comment")) {
        delete((XComment)child);
      } else
        ; // TBD - Need to raise XModelException ?
      deletePCSequence(xdoc,child);
    } // end for
    
    // Remove namespaces
    XNamespace xns = null; 
    for (int i=0; i<namespaces.size(); i++) {
      xns = (XNamespace)namespaces.elementAt(i);
      delete(xns);
    }
  }

  protected void insert(XElement xel) throws SQLException { // returns xelement OID
    long oid = getNewXElementOID();
    xel.setOID(oid);
    String nsURI = xel.getNamespaceURI(); // this may be null
    String localName = xel.getLocalName();
    //System.out.println("inserting element:" + localName);
    String baseURI = xel.getBaseURI();
    String sql = 
      "INSERT INTO X_ELEMENT (X_ELEMENT_OID,NAMESPACE_URI,LOCAL_NAME,BASE_URI) VALUES (?,?,?,?)";
    PreparedStatement pstmt = connection.prepareStatement(sql);
    pstmt.setLong(1,oid);
    pstmt.setString(2,nsURI);
    pstmt.setString(3,localName);
    pstmt.setString(4,baseURI);
    pstmt.executeUpdate();
    pstmt.close();

    // Setting relationships with children
    Vector children = xel.getChildren();
    XNode child;
    for (int i=0; i<children.size(); i++) {
      child = (XNode)children.elementAt(i);
      if (child.getNodeKind().equals("element")) {
        insert((XElement)child);
      } else if (child.getNodeKind().equals("processing-instruction")) {
        insert((XProcessingInstruction)child);
      } else if (child.getNodeKind().equals("comment")) {
        insert((XComment)child);
      } else if (child.getNodeKind().equals("text")) {
        insert((XText)child);
      } else
        // TBD - Raise XModelException ?
        ;
      insertPCSequence(xel,child);
    } // end for

    // Adding attributes
    Vector xattrs = xel.getAttributes();
    XAttribute xattr;
    for (int i=0; i<xattrs.size(); i++) {
      xattr = (XAttribute)xattrs.elementAt(i);
      insert(xattr);
      insertATTRSequence(xel,xattr);
    }

    // Adding namespaces
    Vector xnss = xel.getNamespaces();
    XNamespace xns;
    for (int i=0; i<xnss.size(); i++) {
      xns = (XNamespace)xnss.elementAt(i);
      insert(xns);
      insertNSSequence(xel,xns);
    }
  }

  private void delete(XElement xel) throws SQLException { // returns xelement OID
    long oid = xel.getOID();
    String sql = 
      "DELETE FROM X_ELEMENT WHERE X_ELEMENT_OID = ?";
    PreparedStatement pstmt = connection.prepareStatement(sql);
    pstmt.setLong(1,oid);
    pstmt.executeUpdate();
    pstmt.close();

    // Deleting the children
    Vector children = xel.getChildren();
    XNode child;
    for (int i=0; i<children.size(); i++) {
      child = (XNode)children.elementAt(i);
      if (child.getNodeKind().equals("element")) {
        delete((XElement)child);
      } else if (child.getNodeKind().equals("processing-instruction")) {
        delete((XProcessingInstruction)child);
      } else if (child.getNodeKind().equals("comment")) {
        delete((XComment)child);
      } else if (child.getNodeKind().equals("text")) {
        delete((XText)child);
      } else
        // TBD - Raise XModelException ?
        ;
      deletePCSequence(xel,child);
    } // end for

    Vector xattrs = xel.getAttributes();
    XAttribute xattr;
    for (int i=0; i<xattrs.size(); i++) {
      xattr = (XAttribute)xattrs.elementAt(i);
      delete(xattr);
      deleteATTRSequence(xel,xattr);
    }

    Vector xnss = xel.getNamespaces();
    XNamespace xns;
    for (int i=0; i<xnss.size(); i++) {
      xns = (XNamespace)xnss.elementAt(i);
      deleteNSSequence(xel,xns);
      
      if (!namespaces.contains(xns)) // been here alreay
        namespaces.addElement(xns); // delete namespace later
    }
  }

  private void insert(XAttribute xattr) throws SQLException {
    long oid = getNewXAttributeOID();
    xattr.setOID(oid);
    String nsURI = xattr.getNamespaceURI(); // this may be null
    String localName = xattr.getLocalName();
    String stringValue = xattr.getStringValue();
    String sql = 
      "INSERT INTO X_ATTRIBUTE (X_ATTRIBUTE_OID,NAMESPACE_URI,LOCAL_NAME,STRING_VALUE) VALUES (?,?,?,?)";
    PreparedStatement pstmt = connection.prepareStatement(sql);
    pstmt.setLong(1,oid);
    pstmt.setString(2,nsURI);
    pstmt.setString(3,localName);
    pstmt.setString(4,stringValue);
    pstmt.executeUpdate();
    pstmt.close();

    // Relationship with parent has been already set by XElement
  }

  private void delete(XAttribute xattr) throws SQLException {
    long oid = xattr.getOID();
    xattr.setOID(oid);
    String sql = 
      "DELETE FROM X_ATTRIBUTE WHERE X_ATTRIBUTE_OID = ?";
    PreparedStatement pstmt = connection.prepareStatement(sql);
    pstmt.setLong(1,oid);
    pstmt.executeUpdate();
    pstmt.close();
  }
  
  // TBD - Insert only if the namespace is not found from the table.
  // Set the oid of 'xns' if the namespace is in the table.
  private void insert(XNamespace xns) throws SQLException {
    if (namespaces.contains(xns)) // been here alreay
      return;
    else
      namespaces.addElement(xns);

    long oid = getNewXNamespaceOID();
    xns.setOID(oid);
    String prefix = xns.getPrefix();
    String uri = xns.getURI();
    String sql = 
      "INSERT INTO X_NAMESPACE (X_NAMESPACE_OID,LOCAL_NAME,STRING_VALUE) VALUES (?,?,?)";
    PreparedStatement pstmt = connection.prepareStatement(sql);
    pstmt.setLong(1,oid);
    pstmt.setString(2,prefix);
    pstmt.setString(3,uri);
    pstmt.executeUpdate();
    pstmt.close();
  }

  private void delete(XNamespace xns) throws SQLException {
    long oid = xns.getOID();
    String sql = 
      "DELETE FROM X_NAMESPACE WHERE X_NAMESPACE_OID = ?";
    PreparedStatement pstmt = connection.prepareStatement(sql);
    pstmt.setLong(1,oid);
    pstmt.executeUpdate();
    pstmt.close();
  }
  
  protected void insert(XComment xcomm) throws SQLException {
    long oid = getNewXNamespaceOID();
    xcomm.setOID(oid);
    String stringValue = xcomm.getStringValue();
    String sql = 
      "INSERT INTO X_COMMENT (X_COMMENT_OID,STRING_VALUE) VALUES (?,?)";
    PreparedStatement pstmt = connection.prepareStatement(sql);
    pstmt.setLong(1,oid);
    pstmt.setString(2,stringValue);
    pstmt.executeUpdate();
    pstmt.close();
  }

  private void delete(XComment xcomm) throws SQLException {
    long oid = xcomm.getOID();
    String sql = 
      "DELETE FROM X_COMMENT WHERE X_COMMENT_OID = ?";
    PreparedStatement pstmt = connection.prepareStatement(sql);
    pstmt.setLong(1,oid);
    pstmt.executeUpdate();
    pstmt.close();
  }
  
  private void insert(XText xtext) throws SQLException {
    long oid = getNewXTextOID();
    xtext.setOID(oid);
    String stringValue = xtext.getStringValue();
    String sql = 
      "INSERT INTO X_TEXT (X_TEXT_OID,STRING_VALUE) VALUES (?,?)";
    PreparedStatement pstmt = connection.prepareStatement(sql);
    pstmt.setLong(1,oid);
    pstmt.setString(2,stringValue);
    pstmt.executeUpdate();
    pstmt.close();
  }
  
  private void delete(XText xtext) throws SQLException {
    long oid = xtext.getOID();
    String sql = 
      "DELETE FROM X_TEXT WHERE X_TEXT_OID = ?";
    PreparedStatement pstmt = connection.prepareStatement(sql);
    pstmt.setLong(1,oid);
    pstmt.executeUpdate();
    pstmt.close();
  }
  
  protected void insert(XProcessingInstruction xpi) throws SQLException {
    long oid = getNewXPIOID();
    xpi.setOID(oid);
    String name = xpi.getName();
    String stringValue = xpi.getStringValue();
    String sql = 
      "INSERT INTO X_PI (X_PI_OID,LOCAL_NAME,STRING_VALUE) VALUES (?,?,?)";
    PreparedStatement pstmt = connection.prepareStatement(sql);
    pstmt.setLong(1,oid);
    pstmt.setString(2,name);
    pstmt.setString(3,stringValue);
    pstmt.executeUpdate();
    pstmt.close();
  }

  private void delete(XProcessingInstruction xpi) throws SQLException {
    long oid = xpi.getOID();
    String sql = 
      "DELETE FROM X_PI WHERE X_PI_OID = ?";
    PreparedStatement pstmt = connection.prepareStatement(sql);
    pstmt.setLong(1,oid);
    pstmt.executeUpdate();
    pstmt.close();
  }

  // OID generation
  protected long getNewXDocumentOID() throws SQLException {
    return oidManager.getUniqueOID("document");
  }

  private long getNewXElementOID() throws SQLException {
    return oidManager.getUniqueOID("element");
  }

  private long getNewXAttributeOID() throws SQLException {
    return oidManager.getUniqueOID("attribute");
  }

  private long getNewXNamespaceOID() throws SQLException {
    return oidManager.getUniqueOID("namespace");
  }

  private long getNewXCommentOID() throws SQLException {
    return oidManager.getUniqueOID("comment");
  }

  private long getNewXTextOID() throws SQLException {
    return oidManager.getUniqueOID("text");
  }

  private long getNewXPIOID() throws SQLException {
    return oidManager.getUniqueOID("processing-instruction");
  }

  // Establish parent-child associationship
  protected void insertPCSequence(XNode parent, XNode child) 
      throws SQLException {
    String sql =  "INSERT INTO PC_SEQUENCE VALUES (?,?,?,?)";

    // Performance enhancement point.
    // Make pstmt below an instance variable for reuse.
    PreparedStatement pstmt = connection.prepareStatement(sql);
    pstmt.setLong(1,parent.getOID());
    pstmt.setString(2,parent.getNodeKind());
    pstmt.setLong(3,child.getOID());
    pstmt.setString(4,child.getNodeKind());
    pstmt.executeUpdate();
    pstmt.close();
  }

  private void deletePCSequence(XNode parent, XNode child) 
      throws SQLException {
    String sql =  "DELETE FROM PC_SEQUENCE WHERE PARENT_NODE_OID = ? AND PARENT_NODE_KIND = ? AND CHILD_NODE_OID = ? AND CHILD_NODE_KIND = ?";

    PreparedStatement pstmt = connection.prepareStatement(sql);
    pstmt.setLong(1,parent.getOID());
    pstmt.setString(2,parent.getNodeKind());
    pstmt.setLong(3,child.getOID());
    pstmt.setString(4,child.getNodeKind());
    pstmt.executeUpdate();
    pstmt.close();
  }
  
  // Establish element-attribute associationship
  private void insertATTRSequence(XElement xel, XAttribute xattr) 
     throws SQLException {
    String sql =  "INSERT INTO ATTR_SEQUENCE VALUES (?,?)";

    // Performance enhancement point.
    // Make pstmt below an instance variable for reuse.
    PreparedStatement pstmt = connection.prepareStatement(sql);
    pstmt.setLong(1,xel.getOID());
    pstmt.setLong(2,xattr.getOID());
    pstmt.executeUpdate();
    pstmt.close();
  }

  private void deleteATTRSequence(XElement xel, XAttribute xattr) 
     throws SQLException {
    String sql =  "DELETE FROM ATTR_SEQUENCE WHERE X_ELEMENT_OID = ? AND X_ATTRIBUTE_OID = ?";

    PreparedStatement pstmt = connection.prepareStatement(sql);
    pstmt.setLong(1,xel.getOID());
    pstmt.setLong(2,xattr.getOID());
    pstmt.executeUpdate();
    pstmt.close();
  }

  // Establish element-namespace associationship
  private void insertNSSequence(XElement xel, XNamespace xns) 
     throws SQLException {
    String sql =  "INSERT INTO NS_SEQUENCE VALUES (?,?)";

    // Performance enhancement point.
    // Make pstmt below an instance variable for reuse.
    PreparedStatement pstmt = connection.prepareStatement(sql);
    pstmt.setLong(1,xel.getOID());
    pstmt.setLong(2,xns.getOID());
    pstmt.executeUpdate();
    pstmt.close();
  }

  private void deleteNSSequence(XElement xel, XNamespace xns) 
     throws SQLException {
    String sql =  "DELETE FROM NS_SEQUENCE WHERE X_ELEMENT_OID = ? AND X_NAMESPACE_OID = ?";

    PreparedStatement pstmt = connection.prepareStatement(sql);
    pstmt.setLong(1,xel.getOID());
    pstmt.setLong(2,xns.getOID());
    pstmt.executeUpdate();
    pstmt.close();
  }

  // Handles QName
/*
  private boolean hasPrefix(String qname) {
    if (qname.indexOf(':') != -1)
      return true;
    else
      return false;
  }

  private String getPrefix(String qname) {
    if (hasPrefix(qname)) 
      return qname.substring(0,qname.indexOf(':') - 1);
    else
      return null;
  }

  private String getLocalName(String qname) {
    if (hasPrefix(qname))
      return qname.substring(qname.indexOf(':') + 1);
    else
      return qname;
  }
*/

}
