/*
 * (C) Copyright Bora Software Inc. 2002, 2003 - All Rights Reserved.
 *
 * $Id: SAXToModel.java,v 1.7 2002/10/20 03:01:35 kboo Exp $
 */

package com.bs.xdbms.persistence;

import com.bs.xdbms.datamodel.*;

import java.util.*;

import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.io.PrintWriter;
import java.io.UnsupportedEncodingException;

import org.xml.sax.Attributes;
//import org.xml.sax.Parser;
import org.xml.sax.SAXException;
import org.xml.sax.SAXNotRecognizedException;
import org.xml.sax.SAXNotSupportedException;
import org.xml.sax.SAXParseException;
import org.xml.sax.XMLReader;
import org.xml.sax.ext.LexicalHandler;
import org.xml.sax.helpers.DefaultHandler;
import org.xml.sax.helpers.ParserAdapter;
//import org.xml.sax.helpers.ParserFactory;
import org.xml.sax.helpers.XMLReaderFactory;

// Loads the data model from an input XML file so that the data model
// can be stored to the database subsequently.
// Template adopted from Writer.java

public class SAXToModel extends DefaultHandler implements LexicalHandler {

  // feature ids

  private static final String NAMESPACES_FEATURE_ID = 
    "http://xml.org/sax/features/namespaces";
  private static final String NAMESPACE_PREFIXES_FEATURE_ID = 
    "http://xml.org/sax/features/namespace-prefixes";
  private static final String VALIDATION_FEATURE_ID = 
    "http://xml.org/sax/features/validation";
  private static final String SCHEMA_VALIDATION_FEATURE_ID = 
    "http://apache.org/xml/features/validation/schema";
  private static final String SCHEMA_FULL_CHECKING_FEATURE_ID = 
    "http://apache.org/xml/features/validation/schema-full-checking";

  // property ids

  private static final String LEXICAL_HANDLER_PROPERTY_ID = 
    "http://xml.org/sax/properties/lexical-handler";

  // default settings

  private static final String DEFAULT_PARSER_NAME = 
    "org.apache.xerces.parsers.SAXParser";
  private static final boolean DEFAULT_NAMESPACES = true;
  private static final boolean DEFAULT_VALIDATION = false;
  private static final boolean DEFAULT_SCHEMA_VALIDATION = false;
  private static final boolean DEFAULT_SCHEMA_FULL_CHECKING = false;
  private static final boolean DEFAULT_CANONICAL = false;

  private PrintWriter fOut;
  
  private int fElementDepth;
  private Stack xelStack = new Stack();
  private XDocument xdoc;

  private String baseURI = null;
  private Hashtable xnamespaces = new Hashtable();

  public SAXToModel() {
  }

  public XDocument parse(String uri) 
      throws SAXException, SAXParseException, Exception {
    // uri is not null
    baseURI = uri;
    XMLReader parser = XMLReaderFactory.createXMLReader(DEFAULT_PARSER_NAME);

    // set features
    
    parser.setFeature(NAMESPACE_PREFIXES_FEATURE_ID, false); // This way
    // xmlns properties are handled properly.

    parser.setFeature(NAMESPACES_FEATURE_ID, true);
    parser.setFeature(VALIDATION_FEATURE_ID, false);

    // debugging
    try {
      setOutput(System.out, "UTF8");
    }
    catch (UnsupportedEncodingException e) {
      System.err.println("error: Unable to set output.");
    }

    parser.setContentHandler(this);
    parser.setErrorHandler(this);
    try {
      parser.setProperty(LEXICAL_HANDLER_PROPERTY_ID, this);
    }
    catch (SAXException e) {
      e.printStackTrace();
      // ignore
    }

    // parse file
    parser.parse(uri);

    return xdoc;
  }

  public void setOutput(OutputStream stream, String encoding)
      throws UnsupportedEncodingException {
    if (encoding == null) {
      encoding = "UTF8";
    }

    java.io.Writer writer = new OutputStreamWriter(stream, encoding);
    fOut = new PrintWriter(writer);
  }

  public void setOutput(java.io.Writer writer) {
    fOut = writer instanceof PrintWriter
             ? (PrintWriter)writer : new PrintWriter(writer);
  }

  // ContentHandler methods

  public void startDocument() throws SAXException {
    fElementDepth = 0;
    xdoc = new XDocument();
    xdoc.setBaseURI(baseURI);
  }

  public void processingInstruction(String target, String data)
      throws SAXException {
    try {
      XProcessingInstruction xpi = 
        xdoc.createProcessingInstruction(target,data);
      if (fElementDepth == 0)
        xdoc.appendChild(xpi);
      else
        ((XElement)xelStack.peek()).appendChild(xpi); 
      // TBD - Can xpi be a child of the other node kinds?
    } catch (XModelException xe) {
      throw new SAXException(xe.getMessage());
    }
  }

  public void startElement
                (String uri, 
                 String local, 
                 String raw,
                 Attributes attrs) 
      throws SAXException {
    fOut.println("Entering startElement: " + raw);
    fOut.println("uri is: " + uri);
    fOut.println("localName is: " + local);
    fOut.flush();
    try {
      XElement xel = xdoc.createElement(uri,local);
      xel.setBaseURI(baseURI);
      if (fElementDepth == 0)
        xdoc.appendChild(xel);
      else
        ((XElement)xelStack.peek()).appendChild(xel); 

      xelStack.push(xel);
      fElementDepth++;

      if (attrs != null) {
        XAttribute xattr = null;
        String attrURI = null;
        String attrLocalName = null;
        String attrValue = null;
        int len = attrs.getLength();
        for (int i = 0; i < len; i++) {
          attrURI = attrs.getURI(i);
          attrLocalName = attrs.getLocalName(i);
          attrValue = attrs.getValue(i);

          //System.out.println("attrNsURI is: " + attrURI);
          //System.out.println("attrLocalName is: " + attrLocalName);
          //System.out.println("attrValue is: " + attrValue);

	        xattr = xdoc.createAttribute(attrURI,attrLocalName,attrValue);
          xel.addAttribute(xattr);
        } // end for
      } // if

      // Add a namespace
      XNamespace xns = getNamespace(uri,raw);
      if (xns != null)
        xel.addNamespace(xns);
    } catch (XModelException xe) {
      throw new SAXException(xe.getMessage());
    }
  } 

  // uri is the namespace, raw is the qualified name. ex) p:part
  // Returns null if the prefix is null or the uri is an empty string.
  private XNamespace getNamespace(String uri, String raw) {
    String prefix = getPrefix(raw);
    //fOut.println("Namespace: " + prefix + "=" + uri);
    //fOut.flush();
    if (prefix == null || uri.length() == 0)
      return null;
    else if (xnamespaces.containsKey(prefix)) { // namespaces is Hashtable
      return (XNamespace)xnamespaces.get(prefix);
    }
    else {
      XNamespace xns = new XNamespace(prefix,uri);
      xnamespaces.put(prefix,uri);
      return xns;
    }
  }

  private String getPrefix(String raw) {
    if (raw.indexOf(":") == -1)
      return null; // no prefix
    else
      return raw.substring(0,raw.indexOf(":"));
  }

  /** Characters. */
  public void characters(char ch[], int start, int length)
      throws SAXException {
    //fOut.println("VALUE: " + new String(ch,start,length));
    //fOut.flush();
    try {
      XText xtext = xdoc.createText(new String (ch,start,length));
      ((XElement)xelStack.peek()).appendChild(xtext);
    } catch (XModelException xe) {
      throw new SAXException(xe.getMessage());
    }
  } // characters(char[],int,int);

  /** Ignorable whitespace. */
  public void ignorableWhitespace(char ch[], int start, int length)
      throws SAXException {
    //characters(ch, start, length);
    // TBD
    fOut.println("TBD - Handle ignorableWhitespace()");
    fOut.flush();
  } // ignorableWhitespace(char[],int,int);

  /** End element. */
  public void endElement(String uri, String local, String raw)
      throws SAXException {
    fOut.println("Entering endElement: " + raw);
    fOut.flush();
    xelStack.pop();
    fElementDepth--;
  } // endElement(String)

  //
  // ErrorHandler methods
  //

  /** Warning. */
  public void warning(SAXParseException ex) throws SAXException {
    printError("Warning", ex);
  } // warning(SAXParseException)

  /** Error. */
  public void error(SAXParseException ex) throws SAXException {
    printError("Error", ex);
  } // error(SAXParseException)

  /** Fatal error. */
  public void fatalError(SAXParseException ex) throws SAXException {
    printError("Fatal Error", ex);
    throw ex;
  } // fatalError(SAXParseException)

  //
  // LexicalHandler methods
  //

  /** Start DTD. */
  public void startDTD(String name, String publicId, String systemId)
      throws SAXException {
  } // startDTD(String,String,String)

  /** End DTD. */
  public void endDTD() throws SAXException {
  } // endDTD()

  /** Start entity. */
  public void startEntity(String name) throws SAXException {
  } // startEntity(String)

  /** End entity. */
  public void endEntity(String name) throws SAXException {
  } // endEntity(String)

  /** Start CDATA section. */
  public void startCDATA() throws SAXException {
  } // startCDATA()

  /** End CDATA section. */
  public void endCDATA() throws SAXException {
  } // endCDATA()

  /** Comment. */
  public void comment(char ch[], int start, int length) throws SAXException {
    String value = new String(ch,start,length);
    try {
      XComment xcomm = xdoc.createComment(value);
      if (fElementDepth == 0)
        xdoc.appendChild(xcomm);
      else
        ((XElement)xelStack.peek()).appendChild(xcomm); 
    } catch (XModelException xe) {
      throw new SAXException(xe.getMessage());
    }
  } // comment(char[],int,int)

  //
  // Protected methods
  //

  /** Normalizes and prints the given string. */
  private void normalizeAndPrint(String s) {
    int len = (s != null) ? s.length() : 0;
    for (int i = 0; i < len; i++) {
      char c = s.charAt(i);
      normalizeAndPrint(c);
    }
  } // normalizeAndPrint(String)

  /** Normalizes and prints the given array of characters. */
  private void normalizeAndPrint(char[] ch, int offset, int length) {
    for (int i = 0; i < length; i++) {
      normalizeAndPrint(ch[offset + i]);
    }
  } // normalizeAndPrint(char[],int,int)

    /** Normalizes and print the given character. */
    private void normalizeAndPrint(char c) {

        switch (c) {
            case '<': {
                fOut.print("&lt;");
                break;
            }
            case '>': {
                fOut.print("&gt;");
                break;
            }
            case '&': {
                fOut.print("&amp;");
                break;
            }
            case '"': {
                fOut.print("&quot;");
                break;
            }
            case '\r':
            case '\n': {
                /*
                if (fCanonical) {
                    fOut.print("&#");
                    fOut.print(Integer.toString(c));
                    fOut.print(';');
                    break;
                }
                */
                // else, default print char
            }
            default: {
                fOut.print(c);
            }
        }

    } // normalizeAndPrint(char)

    /** Prints the error message. */
    private void printError(String type, SAXParseException ex) {

        System.err.print("[");
        System.err.print(type);
        System.err.print("] ");
        String systemId = ex.getSystemId();
        if (systemId != null) {
            int index = systemId.lastIndexOf('/');
            if (index != -1)
                systemId = systemId.substring(index + 1);
            System.err.print(systemId);
        }
        System.err.print(':');
        System.err.print(ex.getLineNumber());
        System.err.print(':');
        System.err.print(ex.getColumnNumber());
        System.err.print(": ");
        System.err.print(ex.getMessage());
        System.err.println();
        System.err.flush();

    } // printError(String,SAXParseException)

    //
    // Main
    //

    /** Main program entry point. */
    public static void main(String argv[]) {

        // is there anything to do?
        if (argv.length == 0) {
            printUsage();
            System.exit(1);
        }

        // variables
        SAXToModel writer = null;
        XMLReader parser = null;
        boolean namespaces = DEFAULT_NAMESPACES;
        boolean validation = DEFAULT_VALIDATION;
        boolean schemaValidation = DEFAULT_SCHEMA_VALIDATION;
        boolean schemaFullChecking = DEFAULT_SCHEMA_FULL_CHECKING;
        boolean canonical = DEFAULT_CANONICAL;

        // process arguments
        for (int i = 0; i < argv.length; i++) {
            String arg = argv[i];
            if (arg.startsWith("-")) {
                String option = arg.substring(1);
                if (option.equals("p")) {
                    // get parser name
                    if (++i == argv.length) {
                        System.err.println("error: Missing argument to -p option.");
                    }
                    String parserName = argv[i];

                    // create parser
                    try {
                        parser = XMLReaderFactory.createXMLReader(parserName);
                    }
                    catch (Exception e) {
                      e.printStackTrace();
                      return;
                      /*
                        try {
                            Parser sax1Parser = ParserFactory.makeParser(parserName);
                            parser = new ParserAdapter(sax1Parser);
                            System.err.println("warning: Features and properties not supported on SAX1 parsers.");
                        }
                        catch (Exception ex) {
                            parser = null;
                            System.err.println("error: Unable to instantiate parser ("+parserName+")");
                        }
                    }
                    continue;
                    */
                    }
                }
                if (option.equalsIgnoreCase("n")) {
                    namespaces = option.equals("n");
                    continue;
                }
                if (option.equalsIgnoreCase("v")) {
                    validation = option.equals("v");
                    continue;
                }
                if (option.equalsIgnoreCase("s")) {
                    schemaValidation = option.equals("s");
                    continue;
                }
                if (option.equalsIgnoreCase("f")) {
                    schemaFullChecking = option.equals("f");
                    continue;
                }
                if (option.equalsIgnoreCase("c")) {
                    canonical = option.equals("c");
                    continue;
                }
                if (option.equals("h")) {
                    printUsage();
                    continue;
                }
            }

            // use default parser?
            if (parser == null) {

                // create parser
                try {
                    parser = XMLReaderFactory.createXMLReader(DEFAULT_PARSER_NAME);
                }
                catch (Exception e) {
                    System.err.println("error: Unable to instantiate parser ("+DEFAULT_PARSER_NAME+")");
                    continue;
                }
            }

            // set parser features
            try {
                parser.setFeature(NAMESPACES_FEATURE_ID, namespaces);
            }
            catch (SAXException e) {
                System.err.println("warning: Parser does not support feature ("+NAMESPACES_FEATURE_ID+")");
            }
            try {
                parser.setFeature(VALIDATION_FEATURE_ID, validation);
            }
            catch (SAXException e) {
                System.err.println("warning: Parser does not support feature ("+VALIDATION_FEATURE_ID+")");
            }
            try {
                parser.setFeature(SCHEMA_VALIDATION_FEATURE_ID, schemaValidation);
            }
            catch (SAXNotRecognizedException e) {
                // ignore
            }
            catch (SAXNotSupportedException e) {
                System.err.println("warning: Parser does not support feature ("+SCHEMA_VALIDATION_FEATURE_ID+")");
            }
            try {
                parser.setFeature(SCHEMA_FULL_CHECKING_FEATURE_ID, schemaFullChecking);
            }
            catch (SAXNotRecognizedException e) {
                // ignore
            }
            catch (SAXNotSupportedException e) {
                System.err.println("warning: Parser does not support feature ("+SCHEMA_FULL_CHECKING_FEATURE_ID+")");
            }

            // setup writer
            if (writer == null) {
                writer = new SAXToModel();
                try {
                    writer.setOutput(System.out, "UTF8");
                }
                catch (UnsupportedEncodingException e) {
                    System.err.println("error: Unable to set output. Exiting.");
                    System.exit(1);
                }
            }

            // set parser
            parser.setContentHandler(writer);
            parser.setErrorHandler(writer);
            try {
                parser.setProperty(LEXICAL_HANDLER_PROPERTY_ID, writer);
            }
            catch (SAXException e) {
                // ignore
            }

            // parse file
            try {
                parser.parse(arg);
            }
            catch (SAXParseException e) {
                // ignore
            }
            catch (Exception e) {
                System.err.println("error: Parse error occurred - "+e.getMessage());
                if (e instanceof SAXException) {
                    e = ((SAXException)e).getException();
                }
                e.printStackTrace(System.err);
            }
        }

    } // main(String[])

    //
    // Private static methods
    //

    /** Prints the usage. */
    private static void printUsage() {

        System.err.println("usage: java sax.Writer (options) uri ...");
        System.err.println();

        System.err.println("options:");
        System.err.println("  -p name  Select parser by name.");
        System.err.println("  -n | -N  Turn on/off namespace processing.");
        System.err.println("  -v | -V  Turn on/off validation.");
        System.err.println("  -s | -S  Turn on/off Schema validation support.");
        System.err.println("           NOTE: Not supported by all parsers.");
        System.err.println("  -f  | -F Turn on/off Schema full checking.");
        System.err.println("           NOTE: Requires use of -s and not supported by all parsers.");
        System.err.println("  -c | -C  Turn on/off Canonical XML output.");
        System.err.println("           NOTE: This is not W3C canonical output.");
        System.err.println("  -h       This help screen.");
        System.err.println();

        System.err.println("defaults:");
        System.err.println("  Parser:     "+DEFAULT_PARSER_NAME);
        System.err.print("  Namespaces: ");
        System.err.println(DEFAULT_NAMESPACES ? "on" : "off");
        System.err.print("  Validation: ");
        System.err.println(DEFAULT_VALIDATION ? "on" : "off");
        System.err.print("  Schema:     ");
        System.err.println(DEFAULT_SCHEMA_VALIDATION ? "on" : "off");
        System.err.print("  Schema full checking:     ");
        System.err.println(DEFAULT_SCHEMA_FULL_CHECKING ? "on" : "off");
        System.err.print("  Canonical:  ");
        System.err.println(DEFAULT_CANONICAL ? "on" : "off");

    } // printUsage()

} // class SAXToModel
