/*
 * (C) Copyright Bora Software Inc. 2002, 2003 - All Rights Reserved.
 *
 * $Id: CollectionImpl.java,v 1.18 2002/12/13 22:37:48 kboo Exp $
 */

package com.bs.xdbms.xmldb.base;

import java.util.Hashtable;
import java.util.Vector;
import java.util.Enumeration;
import java.sql.*;

import com.bs.xdbms.datamodel.XDocument;
import com.bs.xdbms.datamodel.XModelException;
import com.bs.xdbms.persistence.DOMToModel;
import com.bs.xdbms.persistence.ModelToDOM;
import com.bs.xdbms.persistence.ModelLoader;
import com.bs.xdbms.persistence.DBUpdater;
import com.bs.xdbms.util.*;

import com.bs.xdbms.xmldb.base.ConfigurableImpl;
import com.bs.xdbms.xmldb.modules.XMLResourceImpl;
import com.bs.xdbms.xmldb.modules.BinaryResourceImpl;
import com.bs.xdbms.xmldb.modules.CollectionManagementServiceImpl;

import org.xmldb.api.modules.XMLResource;
import org.xmldb.api.base.Collection;
import org.xmldb.api.base.Resource;
import org.xmldb.api.base.Service;
import org.xmldb.api.base.XMLDBException;
import org.xmldb.api.base.ErrorCodes;

import org.w3c.dom.Document;
import org.w3c.dom.Node;

/**
 * A <code>Collection</code> represents a collection of <code>Resource</code>s 
 * stored within an XML
 * database. An XML
 * database MAY expose collections as a hierarchical set of parent and child
 * collections.<p />
 *
 * A <code>Collection</code> provides access to the <code>Resource</code>s
 * stored by the <code>Collection</code> and to <code>Service</code> instances
 * that can operate against the <code>Collection</code> and the 
 * <code>Resource</code>s stored within it. The <code>Service</code> mechanism
 * provides the ability to extend the functionality of a <code>Collection</code>
 * in ways that allows optional functionality to be enabled for the <code>Collection</code>. 
 */
final public class CollectionImpl 
    extends ConfigurableImpl 
    implements Collection {

  private String name; // name of this collection

  private Connection connection;
  private Collection parent = null;

  private Hashtable resourceTable = new Hashtable();
  private Hashtable children = new Hashtable();
  private boolean isOpen = true;
 
  // Given an instance of Document, return its originated XDocument.
  // Temperary until XDocument implements Document. 
  private Hashtable doc2xdoc = new Hashtable();

  private Vector services = new Vector();

  private CollectionImpl() {
  }

  private void init() throws SQLException, ClassNotFoundException {
    services.addElement(new CollectionManagementServiceImpl());
    // TBD - createConnection(userId,password);
    connection = ConnectionManager.createConnection(); // collection is open.
    loadResources(); // load resources of this Collection into resourceTable.
    loadChildren();
  }

  /*
   * Collection is always created through CollectionManagementService where
   * a unique id is assigned using properties. One exceptional case is root 
   * collection that is created during installation and its id is 0 (zero).
   */  
  public CollectionImpl(String collectionName, long oid, Collection parent) 
      throws SQLException, ClassNotFoundException, XMLDBException {
    //System.out.println("Creating a collection instance: " + collectionName);
    name = collectionName;
    setProperty(Constants.XDB_COLLECTION_OID,longToString(oid));
    
    if (parent != null)
      setProperty
        (Constants.XDB_COLLECTION_ABSOLUTE_PATH,
         parent.getProperty(Constants.XDB_COLLECTION_ABSOLUTE_PATH)
         + "/"
         + name);
    else
      setProperty(Constants.XDB_COLLECTION_ABSOLUTE_PATH,"/" + name);
      
    this.parent = parent;
    init();
  }
   
  private void loadResources() {
    ExtendedModelLoader loader = null;
    String[] docNames = null;
    try {
      loader = new ExtendedModelLoader(connection);
      docNames = loader.loadDocumentNames(this);
      for (int i=0; i<docNames.length; i++)
        createResource(docNames[i],XMLResource.RESOURCE_TYPE); // TBD
    } catch (Exception e) {
      // handle exception
      e.printStackTrace();
      loader.commit();
    }
    loader.commit();
  }

  private void loadChildren() {
    try { 
      String sql = "SELECT XDB_COLLECTION_OID, NAME FROM XDB_COLLECTION WHERE PARENT = ?";
      PreparedStatement pstmt = connection.prepareStatement(sql);
      pstmt.setLong(1,getOID());
      ResultSet rs = pstmt.executeQuery();
      long oid = -1;
      String name = null;
      Vector v = new Vector();
      Collection child = null;    
      while (rs.next()) {
        oid = rs.getLong(1);
        name = rs.getString(2);
        child = new CollectionImpl(name,oid,this);
        children.put(name,child);        
      }
      pstmt.close();
    } catch (ClassNotFoundException c) {
      c.printStackTrace();
    } catch (java.sql.SQLException se) {
      try {
        connection.rollback();
      } catch (java.sql.SQLException se2) {
        se2.printStackTrace();
      }
      se.printStackTrace();
    } catch (XMLDBException x) {
      x.printStackTrace();
    }
  }
 
  private long getOID() throws XMLDBException {
    return stringToLong(getProperty(Constants.XDB_COLLECTION_OID));
  }
  
  private String longToString(long l) {
    return (new Long(l)).toString();
  }
  
  private long stringToLong(String s) {
    return (new Long(s)).longValue();
  }
       
   /**
    * Returns the name associated with the Collection instance.
    *
    * @return the name of the object.
    * @exception XMLDBException with expected error codes.<br />
    *  <code>ErrorCodes.VENDOR_ERROR</code> for any vendor
    *  specific errors that occur.<br />
    */
  public String getName() throws XMLDBException {
    return name;
  }
   
   /**
    * Provides a list of all services known to the collection. If no services
    * are known an empty list is returned.
    *
    * @return An array of registered <code>Service</code> implementations.
    * @exception XMLDBException with expected error codes.<br />
    *  <code>ErrorCodes.VENDOR_ERROR</code> for any vendor
    *  specific errors that occur.<br />
    *  <code>ErrorCodes.COLLECTION_CLOSED</code> if the <code>close</code> 
    *  method has been called on the <code>Collection</code><br />
    */
  public Service[] getServices() throws XMLDBException {
    return (Service[])services.toArray();
  }

   /**
    * Returns a <code>Service</code> instance for the requested service name and version. If
    * no <code>Service</code> exists for those parameters a null value is returned.
    *
    * @param name Description of Parameter
    * @param version Description of Parameter
    * @return the Service instance or null if no Service could be found.
    * @exception XMLDBException with expected error codes.<br />
    *  <code>ErrorCodes.VENDOR_ERROR</code> for any vendor
    *  specific errors that occur.<br />
    *  <code>ErrorCodes.COLLECTION_CLOSED</code> if the <code>close</code> 
    *  method has been called on the <code>Collection</code><br />
    */
  public Service getService(String name, String version) throws XMLDBException {
    Service service = null;
    for (int i=0; i<services.size(); i++) {
      service = (Service)services.elementAt(i);
      if (service.getName().equals(name)) {
        service.setCollection(this);
        return service; // support version
      }
    }
    return null; // service not found
  }

   /**
    * Returns the parent collection for this collection or null if no parent
    * collection exists.
    *
    * @return the parent <code>Collection</code> instance.
    * @exception XMLDBException with expected error codes.<br />
    *  <code>ErrorCodes.VENDOR_ERROR</code> for any vendor
    *  specific errors that occur.<br />
    *  <code>ErrorCodes.COLLECTION_CLOSED</code> if the <code>close</code> 
    *  method has been called on the <code>Collection</code><br />
    */
  public Collection getParentCollection() throws XMLDBException {
    if (!isOpen())
      throw new XMLDBException(ErrorCodes.COLLECTION_CLOSED);

    return parent;
  }

   /**
    * Returns the number of child collections under this 
    * <code>Collection</code> or 0 if no child collections exist.
    *
    * @return the number of child collections.
    * @exception XMLDBException with expected error codes.<br />
    *  <code>ErrorCodes.VENDOR_ERROR</code> for any vendor
    *  specific errors that occur.<br />
    *  <code>ErrorCodes.COLLECTION_CLOSED</code> if the <code>close</code> 
    *  method has been called on the <code>Collection</code><br />
    */
  public int getChildCollectionCount() throws XMLDBException {
    return children.size();
  }

   /**
    * Returns a list of collection names naming all child collections
    * of the current collection. If no child collections exist an empty list is
    * returned.
    *
    * @return an array containing collection names for all child
    *      collections.
    * @exception XMLDBException with expected error codes.<br />
    *  <code>ErrorCodes.VENDOR_ERROR</code> for any vendor
    *  specific errors that occur.<br />
    *  <code>ErrorCodes.COLLECTION_CLOSED</code> if the <code>close</code> 
    *  method has been called on the <code>Collection</code><br />
    */
  public String[] listChildCollections() throws XMLDBException {
    Vector ids = new Vector();
    Enumeration enum = children.keys();
    Collection child = null;
    String name = null;
    while (enum.hasMoreElements()) {
      name = (String)enum.nextElement();
      child = (Collection)children.get(name);
      if (child.isOpen()) // check if child has been removed
        //ids.addElement(enum.nextElement());
        ids.addElement(name);
      else
        children.remove(name);
    }

    int size = ids.size();
    String[] idss = new String[size];
    for (int i=0; i<size; i++)
      idss[i] = (String)ids.elementAt(i);
    return idss;
  }

   /**
    * Returns a <code>Collection</code> instance for the requested child collection 
    * if it exists.
    *
    * @param name the name of the child collection to retrieve.
    * @return the requested child collection or null if it couldn't be found.
    * @exception XMLDBException with expected error codes.<br />
    *  <code>ErrorCodes.VENDOR_ERROR</code> for any vendor
    *  specific errors that occur.<br />
    *  <code>ErrorCodes.COLLECTION_CLOSED</code> if the <code>close</code> 
    *  method has been called on the <code>Collection</code><br />
    */
  public Collection getChildCollection(String name) throws XMLDBException {
    return (Collection)children.get(name);
  }
   
   /**
    * Returns the number of resources currently stored in this collection or 0
    * if the collection is empty.
    *
    * @return the number of resource in the collection.
    * @exception XMLDBException with expected error codes.<br />
    *  <code>ErrorCodes.VENDOR_ERROR</code> for any vendor
    *  specific errors that occur.<br />
    *  <code>ErrorCodes.COLLECTION_CLOSED</code> if the <code>close</code> 
    *  method has been called on the <code>Collection</code><br />
    */
  public int getResourceCount() throws XMLDBException {
    return resourceTable.size();
  }

   /**
    * Returns a list of the ids for all resources stored in the collection.
    *
    * @return a string array containing the names for all 
    *  <code>Resource</code>s in the collection.
    * @exception XMLDBException with expected error codes.<br />
    *  <code>ErrorCodes.VENDOR_ERROR</code> for any vendor
    *  specific errors that occur.<br />
    *  <code>ErrorCodes.COLLECTION_CLOSED</code> if the <code>close</code> 
    *  method has been called on the <code>Collection</code><br />
    */
  public String[] listResources() throws XMLDBException {
    Vector ids = new Vector();
    Enumeration enum = resourceTable.keys();
    while (enum.hasMoreElements()) {
      ids.addElement(enum.nextElement());
    }

    int size = ids.size();
    String[] idss = new String[size];
    for (int i=0; i<size; i++)
      idss[i] = (String)ids.elementAt(i);
    return idss;
  }
 


   /**
    * Creates a new empty <code>Resource</code> with the provided id. 
    * The type of <code>Resource</code>
    * returned is determined by the <code>type</code> parameter. The XML:DB API currently 
    * defines "XMLResource" and "BinaryResource" as valid resource types.
    * The <code>id</code> provided must be unique within the scope of the 
    * collection. If 
    * <code>id</code> is null or its value is empty then an id is generated by   
    * calling <code>createId()</code>. The
    * <code>Resource</code> created is not stored to the database until 
    * <code>storeResource()</code> is called.
    *
    * @param id the unique id to associate with the created <code>Resource</code>.
    * @param type the <code>Resource</code> type to create.
    * @return an empty <code>Resource</code> instance.    
    * @exception XMLDBException with expected error codes.<br />
    *  <code>ErrorCodes.VENDOR_ERROR</code> for any vendor
    *  specific errors that occur.<br />
    *  <code>ErrorCodes.UNKNOWN_RESOURCE_TYPE</code> if the <code>type</code>
    *   parameter is not a known <code>Resource</code> type.
    *  <code>ErrorCodes.COLLECTION_CLOSED</code> if the <code>close</code> 
    *  method has been called on the <code>Collection</code><br />
    */
  public Resource createResource(String id, String type) throws XMLDBException {
    // TBD - check if id is null or an empty string
    Resource resource = null;
    if (type.equals(XMLResource.RESOURCE_TYPE))
      resource = new XMLResourceImpl(id,this);
    else if (type.equals("BinaryResource"))
      resource = new BinaryResourceImpl(id,this);
    // TBD - add the scopes for id uniqueness depending on the types
    // TBD - check if id is unique
    resourceTable.put(id,resource);
    return resource;
  }

   /**
    * Removes the <code>Resource</code> from the database.
    *
    * @param res the resource to remove.
    * @exception XMLDBException with expected error codes.<br />
    *  <code>ErrorCodes.VENDOR_ERROR</code> for any vendor
    *  specific errors that occur.<br />
    *  <code>ErrorCodes.INVALID_RESOURCE</code> if the <code>Resource</code> is
    *   not valid.<br />
    *  <code>ErrorCodes.NO_SUCH_RESOURCE</code> if the <code>Resource</code> is
    *   not known to this <code>Collection</code>.
    *  <code>ErrorCodes.COLLECTION_CLOSED</code> if the <code>close</code> 
    *  method has been called on the <code>Collection</code><br />
    */
  public void removeResource(Resource res) throws XMLDBException {
    if (res.getResourceType().equals(XMLResource.RESOURCE_TYPE)) {
      Document doc = (Document)((XMLResource)res).getContentAsDOM();
      //DOMToModel d2m = new DOMToModel(res.getId());
      try {
        //XDocument xdoc = d2m.convert(doc);
        XDocument xdoc = (XDocument)doc2xdoc.get(doc);
        remove(xdoc);
      //} catch (XModelException e) {
        // TBD
      //  throw new XMLDBException
      //    (ErrorCodes.VENDOR_ERROR,"CollectionImpl.removeResource() " + e);
      } catch (Exception e2) {
        //e2.printStackTrace();
        throw new XMLDBException
          (ErrorCodes.VENDOR_ERROR,"CollectionImpl.removeResource() " + e2);
      }
    }
    else
      System.out.println
        ("Not implemented removeResource() for " + res.getResourceType());

    resourceTable.remove(res.getId());
  }

  private void remove(XDocument xdoc) throws SQLException {
    DBUpdater updater = null;
    try {
      updater = new DBUpdater(connection);
      //println("Trying to delete the model from the database tables...");
      updater.delete(xdoc);
    } catch (SQLException e) {
      // handle exception
      e.printStackTrace();
      updater.rollback();
      return;
    }
    updater.commit();
  }

   /**
    * Stores the provided resource into the database. If the resource does not
    * already exist it will be created. If it does already exist it will be
    * updated.
    *
    * @param res the resource to store in the database.
    * @exception XMLDBException with expected error codes.<br />
    *  <code>ErrorCodes.VENDOR_ERROR</code> for any vendor
    *  specific errors that occur.<br />
    *  <code>ErrorCodes.INVALID_RESOURCE</code> if the <code>Resource</code> is
    *   not valid.
    *  <code>ErrorCodes.COLLECTION_CLOSED</code> if the <code>close</code> 
    *  method has been called on the <code>Collection</code><br />
    */
  public void storeResource(Resource res) throws XMLDBException {
    if (res.getResourceType().equals(XMLResource.RESOURCE_TYPE)) {
      Document doc = (Document)((XMLResource)res).getContentAsDOM();
      DOMToModel d2m = new DOMToModel(res.getId());
      try {
        XDocument xdoc = d2m.convert(doc);
        store(xdoc);
      } catch (XModelException e) {
        // TBD
        throw new XMLDBException
          (ErrorCodes.VENDOR_ERROR,"CollectionImpl.storeResource() " + e);
      } catch (Exception e2) {
        //e2.printStackTrace();
        throw new XMLDBException
          (ErrorCodes.VENDOR_ERROR,"CollectionImpl.storeResource() " + e2);
      }
    }
    else
      System.out.println
        ("Not implemented storeResource() for " + res.getResourceType());
  }

  private void store(XDocument xdoc) throws SQLException {
    ExtendedDBUpdater updater = null;
    try {
      updater = new ExtendedDBUpdater(connection);
      //println("Trying to insert the model in the database tables...");
      updater.insert(this,xdoc);
    } catch (SQLException e) {
      // handle exception
      e.printStackTrace();
      updater.rollback();
      return;
    } catch (XMLDBException e2) { // failed in getName()
      // handle exception
      e2.printStackTrace();
      updater.rollback();
      return;
    }
    updater.commit();
  }

   /**
    * Retrieves a <code>Resource</code> from the database. If the 
    * <code>Resource</code> could not be
    * located a null value will be returned.
    *
    * @param id the unique id for the requested resource.
    * @return The retrieved <code>Resource</code> instance.
    * @exception XMLDBException with expected error codes.<br />
    *  <code>ErrorCodes.VENDOR_ERROR</code> for any vendor
    *  specific errors that occur.<br />    
    *  <code>ErrorCodes.COLLECTION_CLOSED</code> if the <code>close</code> 
    *  method has been called on the <code>Collection</code><br />
    */
  public Resource getResource(String id) throws XMLDBException {
    // TBD - check isOpen first
    XDocument xdoc = null;
    Node doc = null;
    try {
      xdoc = load(id); 
      if (xdoc == null) // TBD - resource does not exist
      	return null;
      ModelToDOM m2d = new ModelToDOM();
      doc = m2d.convert(xdoc);
      XMLResource resource = 
        (XMLResource)createResource(id,XMLResource.RESOURCE_TYPE);
      resource.setContentAsDOM(doc);
      doc2xdoc.put(doc,xdoc); // for later retrieval of xdoc
      return resource;
    } catch (SQLException e) {
      // TBD
      e.printStackTrace();
      return null;
    } catch (XModelException e2) {
      e2.printStackTrace();
      return null;
    }
  }

  private XDocument load(String baseURI) 
  		throws SQLException, XModelException, XMLDBException {
    ExtendedModelLoader loader = null;
    XDocument xdoc = null;
//    try {
      loader = new ExtendedModelLoader(connection);
      xdoc = loader.loadDocument(this,baseURI);
//    } catch (Exception e) {
      // handle exception
//      e.printStackTrace();
//      throw e;
//    }
    loader.commit();
    return xdoc;
  }

   /**
    * Creates a new unique ID within the context of the <code>Collection</code>
    *
    * @return the created id as a string.
    * @exception XMLDBException with expected error codes.<br />
    *  <code>ErrorCodes.VENDOR_ERROR</code> for any vendor
    *  specific errors that occur.<br />
    *  <code>ErrorCodes.COLLECTION_CLOSED</code> if the <code>close</code> 
    *  method has been called on the <code>Collection</code><br />
    */
   public String createId() throws XMLDBException {
    throw new XMLDBException(ErrorCodes.VENDOR_ERROR,"Not supported");
   }

   /**
    * Returns true if the  <code>Collection</code> is open false otherwise.
    * Calling the <code>close</code> method on 
    * <code>Collection</code> will result in <code>isOpen</code>
    * returning false. It is not safe to use <code>Collection</code> instances
    * that have been closed.
    *
    * @return true if the <code>Collection</code> is open, false otherwise.
    * @exception XMLDBException with expected error codes.<br />
    *  <code>ErrorCodes.VENDOR_ERROR</code> for any vendor
    *  specific errors that occur.<br />
    */
   public boolean isOpen() throws XMLDBException {
     return isOpen;
   }

   /**
    * Releases all resources consumed by the <code>Collection</code>. 
    * The <code>close</code> method must
    * always be called when use of a <code>Collection</code> is complete. It is
    * not safe to use a  <code>Collection</code> after the <code>close</code>
    * method has been called.
    *
    * @exception XMLDBException with expected error codes.<br />
    *  <code>ErrorCodes.VENDOR_ERROR</code> for any vendor
    *  specific errors that occur.<br />
    */
  public void close() throws XMLDBException {
    if (!isOpen)
      return;
    else {
      resourceTable.clear();
      try {
        connection.close();
      } catch (SQLException se) {
        //
      } finally {
        isOpen = false;
      }
    }
  }
}
