/*
 * (C) Copyright Bora Software Inc. 2002, 2003 - All Rights Reserved.
 *
 * $Id: Database.java,v 1.11 2002/11/29 21:06:30 kboo Exp $
 */

package com.bs.xdbms.xmldb.base;

import java.util.*;
import java.sql.*;

import org.xmldb.api.base.Collection;
import org.xmldb.api.base.XMLDBException;
import org.xmldb.api.base.ErrorCodes;

import com.bs.xdbms.xmldb.base.ConfigurableImpl;
import com.bs.xdbms.util.ConnectionManager;
import com.bs.xdbms.util.Constants;

/**
 * <code>Database</code> is an encapsulation of the database driver 
 * functionality that is
 * necessary to access an XML database. Each vendor must provide their own
 * implmentation of the <code>Database</code> interface. The implementation 
 * is registered with the
 * <code>DatabaseManager</code> to provide access to the resources of the XML database.
 * <p />
 * In general usage client applications should only access <code>Database</code>
 * implementations directly during initialization.
 */
final public class Database 
    extends ConfigurableImpl
    implements org.xmldb.api.base.Database {

  private Hashtable collectionTable = new Hashtable();

   /**
    * Returns the name associated with the Database instance.
    *
    * @return the name of the object.
    * @exception XMLDBException with expected error codes.<br />
    *  <code>ErrorCodes.VENDOR_ERROR</code> for any vendor
    *  specific errors that occur.<br />
    */
  public String getName() throws XMLDBException {
    return "myxdb";
  }
   
   /**
    * Retrieves a <code>Collection</code> instance based on the URI provided 
    * in the <code>uri</code> parameter. The format of the URI is defined in the
    * documentation for DatabaseManager.getCollection().<p/>
    *
    * Authentication is handled via username and password however it is not
    * required that the database support authentication. Databases that do not
    * support authentication MUST ignore the 
    * <code>username</code> and <code>password</code> if those provided are not 
    * null. 
    *
    * @param uri the URI to use to locate the collection.
    * @param password The password to use for authentication to the database or
    *    null if the database does not support authentication.
    * @return A <code>Collection</code> instance for the requested collection or
    *  null if the collection could not be found.
    * @return The <code>Collection</code> instance
    * @exception XMLDBException with expected error codes.<br />
    *  <code>ErrorCodes.VENDOR_ERROR</code> for any vendor
    *  specific errors that occur.<br />
    *  <code>ErrroCodes.INVALID_URI</code> If the URI is not in a valid format. <br />
    *  <code>ErrroCodes.PERMISSION_DENIED</code> If the <code>username</code>
    *    and <code>password</code> were not accepted by the database.
    */
  public Collection getCollection(String uri, String username, String password) 
      throws XMLDBException {
    /*
    System.out.println("Collection.getCollection()");
    System.out.println("uri: " + uri);
    System.out.println("user: " + username);
    System.out.println("password: " + password);
    */
    // TBD - Validate uri.
    // ex) myxdb://localhost/collection_name
    if (collectionTable.contains(uri))
      return (Collection)collectionTable.get(uri);
    else {
      Collection collection = null;
      try {
        if (!acceptsURI(uri)) {
           throw new XMLDBException(ErrorCodes.INVALID_URI);
        }
                  
        // TBD - authentication (for the collection) should be done.
        ConnectionManager.setUserId(username);
        ConnectionManager.setPassword(password);
        
        String collectionName = getCollectionName(uri);
        String[] collectionArray = getCollectionArray(collectionName);

        // Note collectionArray[0] == Constants.ROOT_COLLECTION_NAME
        // Get root collection.
        collection = loadCollection(Constants.ROOT_COLLECTION_NAME,null);
        //collection.setProperty(Constants.USER_ID,username);
        //collection.setProperty(Constants.PASSWORD,password);
        
        long parentOID = 
          stringToLong(collection.getProperty(Constants.XDB_COLLECTION_OID));
        for (int i=1; i<collectionArray.length; i++) {        
          collection = loadCollection(collectionArray[i],collection);
        }
          
        collectionTable.put(uri,collection);
        return collection;
      } catch (SQLException e) {
        // TBD
        e.printStackTrace();
        throw new XMLDBException
          (ErrorCodes.PERMISSION_DENIED,"Failed creating a connection");
      } catch (ClassNotFoundException e2) {
        // TBD
        throw new XMLDBException
          (ErrorCodes.PERMISSION_DENIED,"Could not find a JDBC driver");
      }
    } // end else
  }
  
  private String getCollectionName(String uri) {
  	// precondition: acceptsURI(uri) is true.
  	// uri format: "myxdb://localhost/collection_name"
  	// return "/collection_name"
  	int start = uri.indexOf("//");
  	if (start == -1)
      return uri.substring(uri.indexOf("/"));
    else
      return uri.substring(uri.indexOf("/",start+2));
  }

  // name is in the form /x/y/z
  // returns an array of x,y, and z  
  private String[] getCollectionArray(String name) {
    Vector v = new Vector();
    StringTokenizer st = new StringTokenizer(name,"/");
    while (st.hasMoreTokens())
      v.addElement(st.nextToken());

    int k = v.size(); 
    String[] sa = new String[k];
    for (int i=0; i<k; i++)
      sa[i] = (String)v.elementAt(i);
      
    return sa;
  }

   /**
    * acceptsURI determines whether this <code>Database</code> implementation 
    * can handle the URI. It should return true
    * if the Database instance knows how to handle the URI and false otherwise.
    *
    * @param uri the URI to check for.
    * @return true if the URI can be handled, false otherwise.
    * @exception XMLDBException with expected error codes.<br />
    *  <code>ErrorCodes.VENDOR_ERROR</code> for any vendor
    *  specific errors that occur.<br />
    *  <code>ErrroCodes.INVALID_URI</code> If the URI is not in a valid format. <br />
    */
   public boolean acceptsURI(String uri) throws XMLDBException {
     // TBD
     // uri format: "myxdb://localhost/collection_name"
     //System.out.println("uri is : " + uri);
     if (uri.startsWith(getName()))
       return true;
     else
       return false;
   }

   /**
    * Returns the XML:DB API Conformance level for the implementation. This can
    * be used by client programs to determine what functionality is available to
    * them.
    *
    * @return the XML:DB API conformance level for this implementation.
    * @exception XMLDBException with expected error codes.<br />
    *  <code>ErrorCodes.VENDOR_ERROR</code> for any vendor
    *  specific errors that occur.<br />
    */
   public String getConformanceLevel() throws XMLDBException {
     return "Core Level 0";
   }
   
  private Collection loadCollection(String name, Collection parent) 
      throws SQLException, XMLDBException, ClassNotFoundException {
    Connection connection = ConnectionManager.createConnection();
    String sql = "SELECT XDB_COLLECTION_OID, NAME FROM XDB_COLLECTION ";
    sql += "WHERE NAME = ? AND PARENT = ?";

    long parentOID = -1;    
    if (name.equals(Constants.ROOT_COLLECTION_NAME)
        && parent == null) { // root collection
      parentOID = -1;
    } else if (parent != null) {
      parentOID = stringToLong(parent.getProperty(Constants.XDB_COLLECTION_OID));
    }
              
    PreparedStatement pstmt = connection.prepareStatement(sql);
    pstmt.setString(1,name);
    pstmt.setLong(2,parentOID);
    ResultSet rs = pstmt.executeQuery();
    Collection collection = null;
    long oid = -1;    
    while (rs.next()) {
      oid = rs.getLong(1);
      collection = new CollectionImpl(rs.getString(2),oid,parent);
      
      // Add fully qualified collection name.
      if (parent == null) // root collection
        collection.setProperty
          (Constants.XDB_COLLECTION_ABSOLUTE_PATH,
           "/" + Constants.ROOT_COLLECTION_NAME);
      else
        collection.setProperty
          (Constants.XDB_COLLECTION_ABSOLUTE_PATH,
           parent.getProperty(Constants.XDB_COLLECTION_ABSOLUTE_PATH)
           + "/" 
           + collection.getName());
        
    }
    pstmt.close();
    connection.close();

   return collection;
  }
  
  private String longToString(long l) {
    return (new Long(l)).toString();
  }
  
  private long stringToLong(String s) {
    return (new Long(s)).longValue();
  } 
}

